def deferred_methods():
    return [
        'setImageBlendMode',
        'setImageColorMap',
        'setImageColorRange',
        'setImageColorRangeBounds',
        'setImageComponentVisibility',
        'setImageGradientOpacity',
        'setImageGradientOpacityScale',
        'setImageInterpolationEnabled',
        'setImagePiecewiseFunctionGaussians',
        'setImageVolumeSampleDistance',
        'setLabelImageBlend',
        'setLabelImageLabelNames',
        'setLabelImageLookupTable',
        'setLabelImageWeights',
        'selectLayer',
        'setLayerVisibility',
        'setUnits',
        'setViewMode',
        'setXSlice',
        'setYSlice',
        'setZSlice',
    ]
