ITK_VIEWER_SRC = (
    "https://bafybeid7em5cvblohvq65tbk7wckdl4kbfdwvyqqikfjbjo54wsft5vu4a.on.fleek.co/"
)
PYDATA_SPHINX_HREF = "https://cdn.jsdelivr.net/npm/itk-viewer-bootstrap-ui@0.17.0/dist/bootstrapUIMachineOptions.js.es.js"
MUI_HREF = "https://cdn.jsdelivr.net/npm/itk-viewer-material-ui@0.3.0/dist/materialUIMachineOptions.js.es.js"
